import bpy

cachedata = dict()
tkkey = None


class generic_op:
    bl_options = {"REGISTER","UNDO"}
    @classmethod
    def poll(self,context):
        try:
            assert context.active_object.type == "MESH"
        except:
            return False
        finally:
            return True

    
class MESH_OT_vneighbors_edgewise(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.v2v_by_edge"
    bl_label = "by Edge"
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        next_state = bytearray(meshkey[0])
        if (meshkey == tkkey) and (meshkey in cachedata):
            vert_to_vert_map,prev_state = cachedata[meshkey]
        else:
            vert_to_vert_map = {i:{} for i in range(meshkey[0])}
            for a,b in mesh.edge_keys:
                vert_to_vert_map[a][b] = 1
                vert_to_vert_map[b][a] = 1
            tkkey = meshkey
            prev_state = None
        if not prev_state:
            selected_vert_indices = filter(lambda _:mesh.vertices[_].select,range(len(mesh.vertices)))
        else:
            selected_vert_indices = filter(lambda _:mesh.vertices[_].select and not prev_state[_],range(len(mesh.vertices)))
        for v in selected_vert_indices:
            for neighbor_index in vert_to_vert_map[v]:
                next_state[neighbor_index] = True
        mesh.vertices.foreach_set("select",next_state)
        cachedata[meshkey] = (vert_to_vert_map,next_state)
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class MESH_OT_vneighbors_facewise(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.v2v_facewise"
    bl_label = "by Face - Edge"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        next_state = bytearray(meshkey[0])
        if (meshkey == tkkey) and (meshkey in cachedata):
            vert_to_vert_map = cachedata[meshkey]
        else:
            vert_to_vert_map = {i:{} for i in range(meshkey[0])}
            for a,b in mesh.edge_keys:
                vert_to_vert_map[a][b] = 1
                vert_to_vert_map[b][a] = 1
            tkkey = meshkey
        faces = filter(lambda face:(len(face.vertices)==4) and (face.select == False),mesh.faces)
        for f in faces:
            has = False
            t = set()
            for v in f.vertices:
                if mesh.vertices[v].select:
                    has = True
                    t.update(vert_to_vert_map[v])
            if has:
                for v in f.vertices:
                    if not mesh.vertices[v].select:
                        if v not in t:
                            next_state[v]=1 
        mesh.vertices.foreach_set("select",next_state)
        cachedata[meshkey] = vert_to_vert_map
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class MESH_OT_eneighbors_shared_v_f(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.e2e_evfe"
    bl_label = "by Vert+Face"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        state_mask = bytearray(meshkey[1])
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_to_edges_dict = cachedata
        else:
            edge_key_to_index = {k:i for i,k in enumerate(mesh.edge_keys)}
            edge_to_edges_dict = {i:set() for i in range(len(mesh.edges))}
            for f in mesh.faces:
                fed=[edge_key_to_index[k] for k in f.edge_keys]
                for k in f.edge_keys:
                    edge_to_edges_dict[edge_key_to_index[k]].update(fed)
            tkkey = meshkey
        for e in filter(lambda _:mesh.edges[_].select,edge_to_edges_dict):
            k1 = set(mesh.edges[e].key)
            for n in edge_to_edges_dict[e]:
                k2 = set(mesh.edges[n].key)
                if not k1.isdisjoint(k2):
                    state_mask[n] = True
        for e in mesh.edges:
            e.select ^= state_mask[e.index]
        cachedata[meshkey] = edge_key_to_index
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class MESH_OT_eneighbors_shared_v(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.e2e_eve"
    bl_label = "by Vert"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        bpy.ops.object.mode_set(mode="OBJECT")
        mesh = context.active_object.data
        state_mask = bytearray(len(mesh.edges))
        for e in mesh.edges:
            state_mask[e.index] = mesh.vertices[e.vertices[0]].select ^ mesh.vertices[e.vertices[1]].select
        mesh.edges.foreach_set('select',state_mask)
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class MESH_OT_eneighbors_shared_f(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.e2e_efe"
    bl_label = "by Face"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_to_edges_dict = cachedata
        else:
            edge_key_to_index = {k:i for i,k in enumerate(mesh.edge_keys)}
            edge_to_edges_dict = {i:set() for i in range(len(mesh.edges))}
            for f in mesh.faces:
                fed=[edge_key_to_index[k] for k in f.edge_keys]
                for k in f.edge_keys:
                    edge_to_edges_dict[edge_key_to_index[k]].update(fed)
            tkkey = meshkey
        state_mask,esel = (bytearray(meshkey[1]),bytearray(meshkey[1]))
        mesh.edges.foreach_get('select',esel) 
        for e in filter(lambda _:mesh.edges[_].select,range(meshkey[1])):
            for n in edge_to_edges_dict[e]:
                state_mask[n] = 1
        for e in range(meshkey[1]):
            esel[e] ^= state_mask[e]
        mesh.edges.foreach_set('select',esel)
        cachedata[meshkey] = edge_to_edges_dict
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}

class MESH_OT_eneighbors_shared_f_notv(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.e2e_efnve"
    bl_label = "Lateral Neighbors"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        state_mask = bytearray(meshkey[1])
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_to_face_map,edge_key_to_index = cachedata[meshkey]
        else:   
            edge_key_to_index = {}
            edge_to_face_map = {i:set() for i in range(meshkey[1])}
            for i,k in enumerate(mesh.edge_keys):
                edge_key_to_index[k] = i
            for f in mesh.faces:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            tkkey = meshkey
        selected_edge_indices = filter(lambda _:mesh.edges[_].select,range(meshkey[1]))
        for e in selected_edge_indices:
            for f in edge_to_face_map[e]:
                for k in mesh.faces[f].edge_keys:
                    hasv_in = False
                    for v in mesh.edges[e].key:
                        if v in k:
                            hasv_in = True
                    if hasv_in:
                        continue
                    else:
                        state_mask[edge_key_to_index[k]] = True
        for e in filter(lambda _:state_mask[_],range(meshkey[1])):
            mesh.edges[e].select |= state_mask[e]
        cachedata[meshkey] = (edge_to_face_map,edge_key_to_index)
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}



class MESH_OT_eneighbors_shared_v_notf(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.e2e_evnfe"
    bl_label = "Longitudinal Edges"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        state_mask = bytearray(meshkey[1])
        vstate = bytearray(meshkey[0])
        mesh.vertices.foreach_get('select',vstate)
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_to_face_map,vert_to_vert_map,edge_key_to_index = cachedata[meshkey]
        else:
            edge_key_to_index = {}
            vert_to_vert_map = {i:set() for i in range(meshkey[0])}
            edge_to_face_map = {i:set() for i in range(meshkey[1])}
            for i,k in enumerate(mesh.edge_keys):
                edge_key_to_index[k] = i
                vert_to_vert_map[k[0]].add(k[1])
                vert_to_vert_map[k[1]].add(k[0])
            for f in mesh.faces:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            tkkey = meshkey
        selected_edge_indices = filter(lambda _:mesh.edges[_].select,range(meshkey[1]))
        for e in selected_edge_indices:
            for v in mesh.edges[e].key:
                state_mask[v] ^=1
            for f in edge_to_face_map[e]:
                for v in mesh.faces[f].vertices_raw:
                    vstate[v] = 1
        for v in filter(lambda _:state_mask[_],range(meshkey[1])):
            for n in vert_to_vert_map[v]:
                if not vstate[n] and (n != v):
                    mesh.edges[edge_key_to_index[(min(v,n),max(v,n))]].select = True
        cachedata[meshkey] = (edge_to_face_map,vert_to_vert_map,edge_key_to_index)
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}
 
class MESH_OT_just_the_edges(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.je"
    bl_label = "Just the Edge Selection"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        state_mask = bytearray(meshkey[1])
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_key_to_index = cachedata[meshkey]
        else:
            edge_key_to_index = {k:i for i,k in enumerate(mesh.edge_keys)}
            tkkey = meshkey
        for f in filter(lambda _:mesh.faces[_].select,range(meshkey[2])):
            for k in mesh.faces[f].edge_keys:
                state_mask[edge_key_to_index[k]] = 1
        for e in range(meshkey[1]):
            mesh.edges[e].select ^= state_mask[e]
        cachedata[meshkey] = edge_key_to_index
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}

class MESH_OT_inner_edges(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.ie"
    bl_label = "Inner Edge Selection"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        state_mask = bytearray(meshkey[1])
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_to_face_map = cachedata[meshkey]
        else:
            edge_key_to_index = {k:i for i,k in enumerate(mesh.edge_keys)}
            edge_to_face_map = {i:set() for i in range(meshkey[1])}
            for f in mesh.faces:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            tkkey = meshkey
        for e in filter(lambda _:mesh.edges[_].select,range(meshkey[1])):
            for f in edge_to_face_map[e]:
                if mesh.faces[f].select:
                    state_mask[e] ^=1
        for e in range(meshkey[1]):
            mesh.edges[e].select ^= state_mask[e]
        cachedata[meshkey] = edge_to_face_map
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}



class MESH_OT_fneighbors_shared_e(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.f2f_fef"
    bl_label = "by Edge"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        if (meshkey == tkkey) and (meshkey in cachedata):
            face_to_face_map = cachedata[meshkey]
        else:
            edge_key_to_index = {k:i for i,k in enumerate(mesh.edge_keys)}
            edge_to_face_map = {i:set() for i in range(meshkey[1])}
            for f in mesh.faces:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            face_to_face_map = {i:set() for i in range(meshkey[2])}
            for f in mesh.faces:
                for k in f.edge_keys:
                    face_to_face_map[f.index].update(edge_to_face_map[edge_key_to_index[k]])
            tkkey = meshkey
        mask_state = bytearray(meshkey[2])
        for f in filter(lambda _:mesh.faces[_].select,range(meshkey[2])):
            for n in face_to_face_map[f]:
                mask_state[n] = True
        for f in range(meshkey[2]):
            mesh.faces[f].select ^= mask_state[f]
        cachedata[meshkey] = face_to_face_map
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class MESH_OT_fneighbors_shared_v_note(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.f2f_fvnef"
    bl_label = "by Vert not Edge"
    bl_options = {"REGISTER","UNDO"}
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        if (meshkey == tkkey) and (meshkey in cachedata):
            edge_key_to_index = cachedata[meshkey]
        else:
            edge_key_to_index = {k:i for i,k in enumerate(mesh.edge_keys)}
            tkkey = meshkey
        state_mask = bytearray(meshkey[2])
        face_verts = set()
        for f in filter(lambda _:mesh.faces[_].select,range(meshkey[2])):
            face_verts.update(mesh.faces[f].vertices_raw)
        for f in filter(lambda _:not mesh.faces[_].select,range(meshkey[2])):
            ct = 0
            for v in mesh.faces[f].vertices:
                ct += (v in face_verts)
            if ct == 1:
                state_mask[f] = 1
        mesh.faces.foreach_set('select',state_mask)
        cachedata[meshkey] = edge_key_to_index
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class MESH_OT_conway(generic_op,bpy.types.Operator):
    bl_idname = "neighbors.conway"
    bl_label = "Conway"
    def execute(self,context):
        global cachedata,tkkey
        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices),len(mesh.edges),len(mesh.faces),id(self))
        if (meshkey == tkkey) and (meshkey in cachedata):
            vert_to_face_map = cachedata[meshkey]
        else:
            vert_to_face_map = {i:set() for i in range(meshkey[0])}
            for f in mesh.faces:
                for v in f.vertices_raw:
                    vert_to_face_map[v].add(f.index)
            tkkey = meshkey
        sel = set()
        uns = set()
        F = {i:set() for i in range(meshkey[2])}
        for f in range(meshkey[2]):
            for v in mesh.faces[f].vertices_raw:
                for n in filter(lambda _: mesh.faces[_].select and (_ != f),vert_to_face_map[v]):
                    F[f].add(n)
        for f in F:
            if len(F[f]) == 3:
                sel.add(f)
            elif len(F[f]) != 2:
                uns.add(f)
        for f in range(meshkey[2]):
            if f in sel:
                mesh.faces[f].select = True
            if f in uns:
                mesh.faces[f].select = False
        cachedata[meshkey] = vert_to_face_map
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


def vvmenuitem(self,context):
    self.layout.operator(MESH_OT_vneighbors_edgewise.bl_idname)
    self.layout.operator(MESH_OT_vneighbors_facewise.bl_idname)

def eemenuitem(self,context):
    self.layout.operator(MESH_OT_eneighbors_shared_v_f.bl_idname)
    self.layout.operator(MESH_OT_eneighbors_shared_v.bl_idname)
    self.layout.operator(MESH_OT_eneighbors_shared_f.bl_idname)
    self.layout.operator(MESH_OT_eneighbors_shared_f_notv.bl_idname)
    self.layout.operator(MESH_OT_eneighbors_shared_v_notf.bl_idname)
    self.layout.operator(MESH_OT_just_the_edges.bl_idname)
    self.layout.operator(MESH_OT_inner_edges.bl_idname)

def ffmenuitem(self,context):
    self.layout.operator(MESH_OT_fneighbors_shared_e.bl_idname)
    self.layout.operator(MESH_OT_fneighbors_shared_v_note.bl_idname)
    self.layout.operator(MESH_OT_conway.bl_idname)

def tk_header_draw(self,context):
    
    if context.object:
        if context.object.mode == 'EDIT':
            v,e,f = context.tool_settings.mesh_select_mode
            if v:
                vvmenuitem(self,context)
            elif e:
                eemenuitem(self,context)
            elif f:
                ffmenuitem(self,context)


def register():
    bpy.utils.register_module(__name__)
    bpy.types.VIEW3D_MT_edit_mesh_vertices.append(vvmenuitem)
    bpy.types.VIEW3D_MT_edit_mesh_edges.append(eemenuitem)
    bpy.types.VIEW3D_MT_edit_mesh_faces.append(ffmenuitem)
    bpy.types.VIEW3D_HT_header.append(tk_header_draw)

def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.VIEW3D_MT_edit_mesh_vertices.remove(vvmenuitem)
    bpy.types.VIEW3D_MT_edit_mesh_edges.remove(eemenuitem)
    bpy.types.VIEW3D_MT_edit_mesh_faces.remove(ffmenuitem)
    bpy.types.VIEW3D_HT_header.remove(tk_header_draw)

if __name__ == "__main__":
    register()

